//============================================================================
// CFA-3335 Demonstration code
// LOW_LCD.C: Medium-Level LCD drawing routines.
// Samsung KS0107+KS0108 (S6B0107+S6B0108) LCD Controller
// Atmel ATMega32 processor @ 16MHz
// WinAVR / AVR GCC + Atmel AVR Studio
// Written by Brent A. Crosby
// Copyright 2005, Crystalfontz America, Inc. http://www.crystalfontz.com
//============================================================================
#include <avr/pgmspace.h>
#include "typedefs.h"
#include "utils.h"
#include "low_lcd.h"
#include "08x08fnt.h"
#include "12x16fnt.h"
#include "lcd_draw.h"
//============================================================================
// void put_string(ubyte x, ubyte y, ubyte Field_Width, sbyte flash *input);
//
// put_string() will display the null-terminated string pointed to by input
// at (x,y) in the 08x08 font. It will pad a 1-pixel border to the left and
// top of the text. The 08x08 font already has a 1-pixel border at the
// right and bottom.
//
// Put_String will use a "COPY_PUT", replacing the area occupied by the
// string. This means that you can call put_string() several times and then call
// UpdateLCD(), without having to clear rectangles.
//---------------------------------------------------------------------------

void put_string_P(ubyte x, ubyte y, ubyte Field_Width,const prog_char *input)
  {
  register ubyte
    Terminator_Found;
  register ubyte
    Characters_Placed;
  register ubyte
    this_character;
  register ubyte
    *LCD_Memory;
  register const prog_char
	  *Font_Pointer;
  register ubyte
    column;
  register ubyte
    row;
  WORD_UNION
    Clearing_Mask;
  WORD_UNION
    Pixel_Data;

  //Get the first row of the display character.
  row=y>>3;
  //Calculate the address of the first ubyte in display in LCD_Memory
  LCD_Memory=&display[row][x];

  //Calculate Clearing_Mask, the vertical mask that we will and with
  //LCD_Memory to clear the space before we or in the data from the
  //font. It is 9 pixels.
  Clearing_Mask.as_word=~(0x01FF<<(y&0x07));

  //Clear the first col to the left of the string.
  LCD_Memory[0]&=Clearing_Mask.as_bytes[0];
  if(row<7)
    LCD_Memory[128]&=Clearing_Mask.as_bytes[1];
  LCD_Memory++;

  //Initialize Terminator_Found.
  Terminator_Found=0;
  //Move across the field. We will either put the character or a blank
  //in every position of Field_Width.
  for(Characters_Placed=0;Characters_Placed<Field_Width;Characters_Placed++)
    {
    //If we have not passed the terminator, then get the next
    //character in the string. If we find the terminator,
    //remember that we are out of characters.
    if(!Terminator_Found)
      {
      this_character=*input++;
      if(!this_character)
        {
        Terminator_Found=1;
        this_character=' ';
        }
      }
    else
      this_character=' ';
    //Get a pointer into the font information for this
    //character.
    Font_Pointer=&Font_08x08[(this_character-FONT_08X08_BASE)][0];

    //Write the eight columns of this character.
    for(column=0;column<=7;column++)
      {
      //Clear the correct bits in this row and the next row down.
      LCD_Memory[0]&=Clearing_Mask.as_bytes[0];
      if(row<7)
        LCD_Memory[128]&=Clearing_Mask.as_bytes[1];
      //Get the font data, convert it to a word and shift it down. Leave
      //one blank row of pixels above as a spacer.
      Pixel_Data.as_word=((uword)(pgm_read_byte_near(Font_Pointer++)))<<((y&0x07)+1);
      //Set the correct bits in this row and the next row down.
      LCD_Memory[0]|=Pixel_Data.as_bytes[0];
      if(row<7)
        LCD_Memory[128]|=Pixel_Data.as_bytes[1];
      LCD_Memory++;
      }
    }
  }
//---------------------------------------------------------------------------
void put_string(ubyte x, ubyte y, ubyte Field_Width, const char *input)
  {
  register ubyte
    Terminator_Found;
  register ubyte
    Characters_Placed;
  register ubyte
    this_character;
  register ubyte
    *LCD_Memory;
  register const prog_char
	  *Font_Pointer;
  register ubyte
    column;
  register ubyte
    row;
  WORD_UNION
    Clearing_Mask;
  WORD_UNION
    Pixel_Data;

  //Get the first row of the display character.
  row=y>>3;
  //Calculate the address of the first ubyte in display in LCD_Memory
  LCD_Memory=&display[row][x];

  //Calculate Clearing_Mask, the vertical mask that we will and with
  //LCD_Memory to clear the space before we or in the data from the
  //font. It is 9 pixels.
  Clearing_Mask.as_word=~(0x01FF<<(y&0x07));

  //Clear the first col to the left of the string.
  LCD_Memory[0]&=Clearing_Mask.as_bytes[0];
  if(row<7)
    LCD_Memory[128]&=Clearing_Mask.as_bytes[1];
  LCD_Memory++;

  //Initialize Terminator_Found.
  Terminator_Found=0;
  //Move across the field. We will either put the character or a blank
  //in every position of Field_Width.
  for(Characters_Placed=0;Characters_Placed<Field_Width;Characters_Placed++)
    {
    //If we have not passed the terminator, then get the next
    //character in the string. If we find the terminator,
    //remember that we are out of characters.
    if(!Terminator_Found)
      {
      this_character=*input++;
      if(!this_character)
        {
        Terminator_Found=1;
        this_character=' ';
        }
      }
    else
      this_character=' ';
    //Get a pointer into the font information for this
    //character.
    Font_Pointer=&Font_08x08[(this_character-FONT_08X08_BASE)][0];

    //Write the eight columns of this character.
    for(column=0;column<=7;column++)
      {
      //Clear the correct bits in this row and the next row down.
      LCD_Memory[0]&=Clearing_Mask.as_bytes[0];
      if(row<7)
        LCD_Memory[128]&=Clearing_Mask.as_bytes[1];
      //Get the font data, convert it to a word and shift it down. Leave
      //one blank row of pixels above as a spacer.
      Pixel_Data.as_word=((uword)(pgm_read_byte_near(Font_Pointer++)))<<((y&0x07)+1);
      //Set the correct bits in this row and the next row down.
      LCD_Memory[0]|=Pixel_Data.as_bytes[0];
      if(row<7)
        LCD_Memory[128]|=Pixel_Data.as_bytes[1];
      LCD_Memory++;
      }
    }
  }
//============================================================================
// void invert_string(ubyte x, ubyte y, ubyte Field_Width);
//
// invert_string() will invert an area on the screen starting at (x-1, y-1)
// and extending to (x+Field_Width*9+1, y+9).
//---------------------------------------------------------------------------
void invert_string(ubyte x, ubyte y, ubyte Field_Width)
  {
  register ubyte
    *LCD_Memory;
  register ubyte
    column;
  register ubyte
    row;
  register ubyte
    Columns_To_Invert;
  WORD_UNION
    Inverting_Mask;

  //Get the first row of the display character.
  row=y>>3;
  //Calculate the address of the first ubyte in display in LCD_Memory
  LCD_Memory=&display[row][x];

  //Calculate Inverting_Mask, the vertical mask that we will xor with
  //LCD_Memory to invert the area. It is 9 pixels.
  Inverting_Mask.as_word=0x01FF<<(y&0x07);

  //Invert the first col to the left of the string, and 8 more for
  //each of Field_Width characters.
  Columns_To_Invert=Field_Width<<3;
  for(column=0;column<=Columns_To_Invert;column++)
    {
    //Clear the correct bits in this row and the next row down.
    LCD_Memory[0]^=Inverting_Mask.as_bytes[0];
    if(row<7)
      LCD_Memory[128]^=Inverting_Mask.as_bytes[1];
    LCD_Memory++;
    }
  }
//============================================================================
void horizontal_line(ubyte x1, ubyte y, ubyte x2)
  {
  register ubyte
    *LCD_Memory;
  register ubyte
    column;
  register ubyte
    Set_Mask;

  //Bail for bogus parametrers.
  if((x2<x1)||
     (127<x1)||
     (127<x2)||
     (63<y))
    return;

  //Calculate the address of the first ubyte in display in LCD_Memory
  LCD_Memory=&display[y>>3][x1];

  //Calculate Set_Mask, the vertical mask that we will or with
  //LCD_Memory to clear the space before we or in the data from the
  //font. It is 9 pixels.
  Set_Mask=0x01<<(y&0x07);

  //Move across memory, oring the mask in.
  for(column=x1;column<=x2;column++)
    {
    LCD_Memory[0]|=Set_Mask;
    LCD_Memory++;
    }
  }
//============================================================================
void vertical_line(ubyte x, ubyte y1, ubyte y2)
  {
  register ubyte
    *LCD_Memory_1;
  register ubyte
    *LCD_Memory_2;
  //Bail for bogus parametrers.
  if((y2<y1)||
     (127<x)||
     (63<y1)||
     (63<y2))
    return;
  //Draw the last pixel too.
  y2++;
  //Calculate the address of the first ubyte in display in LCD_Memory
  LCD_Memory_1=&display[y1>>3][x];
  //Calculate the address of the last ubyte in display in LCD_Memory
  LCD_Memory_2=&display[y2>>3][x];
  //If they are the same, then this is a special case.
  if(LCD_Memory_1==LCD_Memory_2)
    //The line fits in one ubyte.
    *LCD_Memory_1|=(0xFF<<(y1&0x07))&(~(0xFF<<(y2&0x07)));
  else
    {
    //This is the case where we need to loop over multiple
    //(possibly 0) lines rows. Do the top ubyte.
    *LCD_Memory_1|=(0xFF<<(y1&0x07));
    //Do the bytes in the middle.
    for(LCD_Memory_1+=128;LCD_Memory_1!=LCD_Memory_2;LCD_Memory_1+=128)
      *LCD_Memory_1=0xFF;
    //Do the bottom ubyte.
    *LCD_Memory_2|=~(0xFF<<(y2&0x07));
    }
  }
//============================================================================
void outline_rectangle(ubyte x1, ubyte y1, ubyte x2, ubyte y2)
  {
  horizontal_line(x1+1,y1,x2-1);
  horizontal_line(x1+1,y2,x2-1);
  vertical_line(x1,y1,y2);
  vertical_line(x2,y1,y2);
  }
//============================================================================
void clear_rectangle(ubyte x1,ubyte y1,ubyte x2,ubyte y2)
  {
  register ubyte
    *LCD_Memory;
  register ubyte
    *LCD_Memory_1;
  register ubyte
    *LCD_Memory_2;
  register ubyte
    mask;
  register ubyte
    column;
  //Draw the last pixel too.
  y2++;
  //Bail for bogus parametrers.
  if((x2<x1)||
     (y2<y1)||
     (127<x1)||
     (127<x2)||
     (63<y1)||
     (63<y2))
    return;
  //Calculate the address of the first ubyte in display in LCD_Memory
  LCD_Memory_1=&display[y1>>3][x1];
  //Calculate the address of the last ubyte in display in LCD_Memory
  LCD_Memory_2=&display[y2>>3][x1];
  //If they are the same, then this is a special case.
  if(LCD_Memory_1==LCD_Memory_2)
    {
    //The rectangle fits in one row.
    mask=~((0xFF<<(y1&0x07))&(~(0xFF<<(y2&0x07))));
    for(column=x1;column<=x2;column++)
      {
      *LCD_Memory_1&=mask;
      LCD_Memory_1++;
      }
    }
  else
    {
    //This is the case where we need to loop over multiple
    //(possibly 0) lines rows. Do the top ubyte.
    mask=~(0xFF<<(y1&0x07));
    LCD_Memory=LCD_Memory_1;
    for(column=x1;column<=x2;column++)
      {
      *LCD_Memory&=mask;
      LCD_Memory++;
      }
    //Do the bytes in the middle.
    for(LCD_Memory_1+=128;LCD_Memory_1!=LCD_Memory_2;LCD_Memory_1+=128)
      {
      LCD_Memory=LCD_Memory_1;
      for(column=x1;column<=x2;column++)
        {
        *LCD_Memory=0;
        LCD_Memory++;
        }
      }
    //Do the bottom ubyte.
    mask=0xFF<<(y2&0x07);
    for(column=x1;column<=x2;column++)
      {
      *LCD_Memory_2&=mask;
      LCD_Memory_2++;
      }
    }
  }
//============================================================================
// Only '0'-'9', '.', '+', '-', '%' and 'e' will be displayed. The font is
// 12 x 16.
void put_big_number(ubyte x, ubyte y, ubyte Field_Width, ubyte *input)
  {
  register ubyte
    Terminator_Found;
  register ubyte
    Characters_Placed;
  register ubyte
    this_character;
  register ubyte
    row;
  register ubyte
    row_limit;
  register ubyte
    *LCD_Memory;
  register const prog_int16_t
	  *Font_Pointer;
  register ubyte
    column;
  DWORD_UNION
    Clearing_Mask;
  DWORD_UNION
    Pixel_Data;

  //We typically need to manipulate 3 rows of data. We only need to do
  //two if we are in the bottom two rows or the top two rows of the display.
  row_limit=y>>3;

  //Calculate the address of the first ubyte in display in LCD_Memory
  LCD_Memory=&display[row_limit][x];

  if((row_limit<2)||(5<row_limit))
    row_limit=2;
  else
    row_limit=3;

  //Calculate Clearing_Mask, the vertical mask that we will and with
  //LCD_Memory to clear the space before we or in the data from the
  //font. It is 16 pixels.
  Clearing_Mask.as_dword=~(0x01FFFFL<<(y&0x07));

  //Clear the first col to the left of the string.
  for(row=0;row<=row_limit;row++)
    LCD_Memory[row*128]&=Clearing_Mask.as_bytes[row];
  LCD_Memory++;

  //Initialize Terminator_Found.
  Terminator_Found=0;
  //Move across the field. We will either put the character or a blank
  //in every position of Field_Width.
  for(Characters_Placed=0;Characters_Placed<Field_Width;Characters_Placed++)
    {
    //If we have not passed the terminator, then get the next
    //character in the string. If we find the terminator,
    //remember that we are out of characters.
    if(!Terminator_Found)
      {
      this_character=*input++;
      if(!this_character)
        {
        Terminator_Found=1;
        this_character=' ';
        }
      }
    else
      this_character=' ';
    //Get a pointer into the font information for this character.
    switch(this_character)
      {
      case '+': Font_Pointer=&big_plus[0]; break;
      case '-': Font_Pointer=&big_minus[0]; break;
      case '.': Font_Pointer=&big_dot[0]; break;
      case '0': Font_Pointer=&big_0[0]; break;
      case '1': Font_Pointer=&big_1[0]; break;
      case '2': Font_Pointer=&big_2[0]; break;
      case '3': Font_Pointer=&big_3[0]; break;
      case '4': Font_Pointer=&big_4[0]; break;
      case '5': Font_Pointer=&big_5[0]; break;
      case '6': Font_Pointer=&big_6[0]; break;
      case '7': Font_Pointer=&big_7[0]; break;
      case '8': Font_Pointer=&big_8[0]; break;
      case '9': Font_Pointer=&big_9[0]; break;
      default: Font_Pointer=&big_space[0];
      }

    //Write the twelve columns of this character.
    for(column=0;column<=11;column++)
      {
      //Clear the correct bits in this row and the next two rows down.
      for(row=0;row<=row_limit;row++)
        LCD_Memory[(uword)row<<7]&=Clearing_Mask.as_bytes[row];
      //Get the font data, convert it to a word and shift it down. Leave
      //one blank row of pixels above as a spacer.
      Pixel_Data.as_dword=((dword)(pgm_read_word_near(Font_Pointer++)))<<(y&0x07);
      //Set the correct bits in this row and the next row down.
      for(row=0;row<=row_limit;row++)
        LCD_Memory[(uword)row<<7]|=Pixel_Data.as_bytes[row];
      LCD_Memory++;
      }
    }
  }
//============================================================================
